# Copyright (c) 2021-2022 Jaakko Keränen <jaakko.keranen@iki.fi>
# License: BSD-2-Clause

import fnmatch
import hashlib
import importlib
import os.path
import select
import socket
import multiprocessing as mp
import threading
import queue
import re
import signal
import time
from pathlib import Path
from urllib.parse import urlparse

import OpenSSL.crypto
from OpenSSL import SSL, crypto


class GeminiError(Exception):
    def __init__(self, status, msg):
        Exception.__init__(self, msg)
        self.status = status


class AbortedIOError(Exception):
    def __init__(self, msg):
        Exception.__init__(self, msg)


def wait_for_read(stream, timeout):
    fno = stream._socket.fileno()
    r, _, x = select.select([fno], [], [fno], timeout)
    if len(x):
        raise AbortedIOError('recv: socket in error state')
    if len(r) == 0:
        raise socket.timeout('stalled: not ready for reading')


def wait_for_write(stream, timeout):
    fno = stream._socket.fileno()
    _, w, x = select.select([], [fno], [fno], timeout)
    if len(x):
        raise AbortedIOError('send: socket in error state')
    if len(w) == 0:
        raise socket.timeout('stalled: not ready for writing')


def safe_recv(stream, max_len, stall_timeout=10):
    data = bytearray()
    remain = max_len
    while remain > 0:
        try:
            incoming = stream.recv(remain)
            remain -= len(incoming)
            data += bytearray(incoming)

            if len(data):
                # Got something, return it asap.
                break

            # Wait until reading is possible.
            wait_for_read(stream, stall_timeout)

        except OpenSSL.SSL.WantReadError:
            wait_for_read(stream, stall_timeout)
        except OpenSSL.SSL.WantWriteError:
            pass
        except OpenSSL.SSL.WantX509LookupError:
            pass
    return data


def is_bytes(data):
    return type(data) == bytes or type(data) == bytearray


def safe_sendall(stream, data, stall_timeout=30):
    """
    Send data over an SSL connection, accounting for stalls and retries
    required by OpenSSL.

    Args:
        stream (OpenSSL.SSL.Connection): Network stream.
        data (bytes or file-like): Data to sent. If not a bytes/bytearray,
            ``read()`` will be called to get more data.
        stall_timeout (float): Number of seconds to wait until
            terminating a stalled send.
    """
    try:
        if is_bytes(data):
            streaming = False
        else:
            streaming = True
            if isinstance(data, Path):
                print('opening', data)
                data = open(data, 'rb')

        # We may need to retry sending with the exact same buffer,
        # so keep it around until successful.
        BUF_LEN = 32768
        if streaming:
            send_buf = data.read(BUF_LEN)
        else:
            send_buf = data[:BUF_LEN]

        last_time = time.time()
        pos = 0
        while len(send_buf) > 0:
            try:
                if time.time() - last_time > stall_timeout:
                    raise AbortedIOError('stalled')
                sent = stream.send(send_buf)
                if sent < 0:
                    raise AbortedIOError('failed to send')
                pos += sent
                if streaming:
                    send_buf = send_buf[sent:]
                    if len(send_buf) < BUF_LEN / 2:
                        send_buf += data.read(BUF_LEN)
                else:
                    send_buf = data[pos : pos + BUF_LEN]
                if sent > 0:
                    last_time = time.time()
                else:
                    wait_for_write(stream, stall_timeout)
            except OpenSSL.SSL.WantReadError:
                pass
            except OpenSSL.SSL.WantWriteError:
                # Wait until the socket is ready for writing.
                wait_for_write(stream, stall_timeout)
            except OpenSSL.SSL.WantX509LookupError:
                pass

    finally:
        # Close resources and handles.
        if data and hasattr(data, 'close'):
            data.close()


def safe_close(stream):
    if not stream:
        return
    try:
        stream.shutdown()
    except Exception as er:
        print('stream shutdown error:', er)
    try:
        stream.close()
    except Exception as er:
        print('stream close error:', er)


def report_error(stream, code, msg):
    print(time.strftime('%Y-%m-%d %H:%M:%S'), f'   ', '--', code, msg)
    #stream.sendall(f'{code} {msg}\r\n'.encode('utf-8'))
    safe_sendall(stream, f'{code} {msg}\r\n'.encode('utf-8'))
    safe_close(stream)


def cert_subject(cert):
    comps = {}
    for name, value in cert.get_subject().get_components():
        comps[name.decode()] = value.decode()
    return comps

def cert_issuer(cert):
    comps = {}
    for name, value in cert.get_issuer().get_components():
        comps[name.decode()] = value.decode()
    return comps


class Identity:
    """
    Client certificate.

    SHA-256 hashes are calculated automatically for the whole certificate and
    just for the public key.

    Attributes:
        cert (bytes): Certificate (DER format).
        pubkey (bytes): Public key (DER format).
        fp_cert (str): SHA-256 hash of the certificate.
        fp_pubkey (str): SHA-256 hash of the public key.
    """
    def __init__(self, cert):
        self._subject = cert_subject(cert)
        self._issuer = cert_issuer(cert)
        self.cert = crypto.dump_certificate(crypto.FILETYPE_ASN1, cert)
        m = hashlib.sha256()
        m.update(self.cert)
        self.fp_cert = m.hexdigest()
        self.pubkey = crypto.dump_publickey(crypto.FILETYPE_ASN1, cert.get_pubkey())
        m = hashlib.sha256()
        m.update(self.pubkey)
        self.fp_pubkey = m.hexdigest()

    def __str__(self):
        return f"{self.fp_cert};{self.fp_pubkey}"

    def subject(self):
        """
        Returns:
            dict: Name components of the certificate subject, e.g.: ``{'CN': 'Name'}``.
        """
        return self._subject

    def issuer(self):
        """
        Returns:
            dict: Name components of the certificate issuer, e.g.: ``{'CN': 'Name'}``.
        """
        return self._issuer


class Request:
    """
    Request received from a client.

    Request objects are used to pass information to entry points handlers.
    One does not need to construct them directly.

    Attributes:
        remote_address (str): IP address of the client.
        scheme (str): Request protocol scheme. Either ``gemini`` or ``titan``.
        identity (gmcapsule.gemini.Identity): Client certificate.
            May be ``None``.
        hostname (str): Hostname.
        path (str): URL path. Always begins with a ``/``.
        query (str): Encoded query string. You can use `urllib.parse.unquote()
            <https://docs.python.org/3/library/urllib.parse.html#urllib.parse.unquote>`_
            to decode the percent-coding. ``None`` if the URL does not have a query
            string.
        content_token (str): Encoded token specified in Titan URLs.
            May be ``None``.
        content_mime (str): MIME type specified in Titan URls. May be ``None``.
        content (bytes): Content uploaded by the client in a Titan request.
            May be ``None``.
    """
    def __init__(self, identity=None, scheme='gemini', hostname='', path='', query=None,
                 remote_address=None, content_token=None, content_mime=None, content=None,
                 worker_id=None):
        self.remote_address = remote_address
        self.scheme = scheme
        self.identity = identity
        self.hostname = hostname
        self.path = path
        self.query = query
        self.content_token = content_token
        self.content_mime = content_mime
        self.content = content
        self.worker_id = worker_id

    def url(self):
        return f'{self.scheme}://{self.hostname}{self.path}{"?" + self.query if self.query else ""}'


def verify_callback(connection, cert, err_num, err_depth, ret_code):
    #print("verify_callback:", connection, cert, ret_code)
    return True


class Cache:
    """
    Response cache base class.

    Derived classes are expected to override the save() and try_load()
    methods to save and load response content as appropriate.

    The server will not try to load or save cached content when a request
    includes a query string or a client certificate is provided. When multiple
    Cache objects have been installed, the save/load operation is attempted
    on each in turn until one cache succeeds in saving or loading content.

    Each server worker thread constructs their own Cache objects. If there
    is a shared backing store like a file system or a database, proper care
    should be taken to synchronize access to it from the Cache objects.

    The mapping from URLs to cache paths is::

        gemini://example.com/path/file.gmi
         ↓
        /example.com/path/file.gmi

    """

    def __init__(self):
        pass

    def save(self, path, media_type, content):
        """
        Save content to the cache.

        Args:
            path (str): URL path being loaded with the hostname prepended as
                the top-level directory.
            media_type (str): MIME type, e.g., "text/plain".
            content (bytes): Content to save.

        Returns:
            bool: True if successfully saved.
        """
        return False

    def try_load(self, path):
        """
        Load content from the cache.

        Args:
            path (str): URL path being loaded with the hostname prepended as
                the top-level directory.

        Returns:
            tuple(str, bytes): Content MIME type and data. Returns
            (None, None) if nothing is cached for the given path.
        """
        return None, None


def handle_gemini_or_titan_request(request_data):
    worker = request_data.worker
    stream = request_data.stream
    data = request_data.buffered_data
    from_addr = request_data.from_addr
    identity = request_data.identity
    request = request_data.request
    expected_size = 0
    req_token = None
    req_mime = None

    if request.startswith('titan:'):
        if identity is None and worker.cfg.require_upload_identity():
            report_error(stream, 60, "Client certificate required for upload")
            return
        # Read the rest of the data.
        parms = request.split(';')
        request = parms[0]
        for p in parms:
            if p.startswith('size='):
                expected_size = int(p[5:])
            elif p.startswith('token='):
                req_token = p[6:]
            elif p.startswith('mime='):
                req_mime = p[5:]
        worker.log(f'Receiving Titan content: {expected_size}')
        max_upload_size = worker.cfg.max_upload_size()
        if expected_size > max_upload_size and max_upload_size > 0:
            report_error(stream, 59, "Maximum content length exceeded")
            return
        while len(data) < expected_size:
            incoming = safe_recv(stream, 65536)
            if len(incoming) == 0:
                break
            data += incoming
        if len(data) != expected_size:
            report_error(stream, 59, "Invalid content length")
            return
    else:
        # No Payload in Gemini.
        if len(data):
            report_error(stream, 59, "Bad request")
            return

    url = urlparse(request)
    path = url.path
    if path == '':
        path = '/'
    hostname = url.hostname

    if url.port != None and url.port != worker.port:
        report_error(stream, 59, "Invalid port number")
        return
    if not stream.get_servername():
        # Server name indication is required.
        report_error(stream, 59, "Missing TLS server name indication")
        return
    if stream.get_servername().decode() != hostname:
        report_error(stream, 53, "Proxy request refused")
        return

    try:
        status, meta, body, from_cache = worker.context.call_entrypoint(Request(
            identity,
            remote_address=from_addr,
            scheme=url.scheme,
            hostname=hostname,
            path=path,
            query=url.query if '?' in request else None,
            content_token=req_token,
            content_mime=req_mime,
            content=data if len(data) else None,
            worker_id=request_data.worker.id
        ))

        output = f'{status} {meta}\r\n'.encode('utf-8')
        if is_bytes(body):
            safe_sendall(stream, output + body)
        else:
            # `body` is some sort of streamable data, cannot send in one call.
            safe_sendall(stream, output)
            safe_sendall(stream, body)

        # Save to cache.
        if not from_cache and status == 20 and is_bytes(body):
            for cache in worker.context.caches:
                if cache.save(hostname + path, meta, body):
                    break

    except GeminiError as error:
        report_error(stream, error.status, str(error))
        return


def unpack_response(response):
    if type(response) == tuple:
        if len(response) == 2:
            status, meta = response
            response = ''
        else:
            status, meta, response = response
    else:
        status = 20
        meta = 'text/gemini; charset=utf-8'

    if response == None:
        body = b''
    elif type(response) == str:
        body = response.encode('utf-8')
    else:
        body = response

    return status, meta, body


class Context:
    def __init__(self, cfg, allow_extension_workers=False, handler_queue=None,
                 response_queues=None):
        self.cfg = cfg
        self.is_quiet = True
        self.allow_extension_workers = allow_extension_workers
        if allow_extension_workers:
            self.shutdown = threading.Event()
        self.hostnames = cfg.hostnames()
        self.entrypoints = {'gemini': {}, 'titan': {}}
        for proto in ['gemini', 'titan']:
            self.entrypoints[proto] = {}
            for hostname in self.hostnames:
                self.entrypoints[proto][hostname] = []
        self.caches = []
        self.protocols = {}
        self.add_protocol('gemini', handle_gemini_or_titan_request)
        self.add_protocol('titan', handle_gemini_or_titan_request)
        # Queue for pending handler jobs.
        self.job_lock = threading.Lock()
        self.job_id = 0
        self.handler_queue = handler_queue
        self.response_queues = response_queues

    def config(self):
        return self.cfg

    def is_background_work_allowed(self):
        """
        Determines whether extension modules are allowed to start workers.
        """
        return self.allow_extension_workers

    def shutdown_event(self):
        """
        Returns:
            threading.Event: Extension modules must check this to be notified
            when the server is being shut down.
        """
        if not self.is_background_work_allowed():
            raise Exception("background work not allowed")
        # This is used in a parser thread that is allowed to launch workers.
        return self.shutdown

    def set_quiet(self, is_quiet):
        self.is_quiet = is_quiet

    def print(self, *args):
        if not self.is_quiet:
            print(*args)

    def add_entrypoint(self, protocol, hostname, path_pattern, entrypoint):
        self.entrypoints[protocol][hostname].append((path_pattern, entrypoint))

    def __setitem__(self, key, value):
        for hostname in self.hostnames:
            self.add_entrypoint('gemini', hostname, key, value)

    def add_cache(self, cache):
        """
        Install a cache.

        All installed caches will attempt to save and load content until one
        succeeds. The caches installed first get precedence.

        Args:
            cache (Cache): Cache instance.
        """
        self.caches.append(cache)

    def add_protocol(self, scheme, handler):
        """
        Registers a new protocol handler.

        Args:
            scheme (str): URL scheme of the protocol.
            handler (callable): Handler to be called when a request with the
                specified scheme is received. The handler must return the
                response to be sent to the client (bytes).
        """
        self.protocols[scheme] = handler

    def add(self, path, entrypoint, hostname=None, protocol='gemini'):
        """
        Register a URL entry point.

        Extension modules must call this to become visible in the server's
        path hierarchy. Entry points are looked up in the order the modules
        were loaded, with earlier modules getting precedence.

        Args:
            path (str): URL path. Must begin with a slash (``/``). Asterisk
                wildcards (``*``) are supported. Note that if the path
                ``/*`` is registered, it will match any requested URL.
            entrypoint (callable): Function or other callable object that
                gets called when a request is processed with a matching
                URL path. A :class:`~gmcapsule.gemini.Request` is passed in as the
                only argument.
            hostname (str): Hostname for the entry point. If omitted,
                the entry point applies to all configured hostnames.
            protocol (str): Protocol for the entry point.
        """
        if hostname:
            self.add_entrypoint(protocol, hostname, path, entrypoint)
        else:
            for hostname in self.cfg.hostnames():
                if not hostname:
                    raise Exception(f'invalid hostname: "{hostname}"')
                self.add_entrypoint(protocol, hostname, path, entrypoint)

    def load_modules(self):
        # The configuration can override default priorities.
        mod_priority = {}
        if 'priority' in self.cfg.ini:
            for name, priority in self.cfg.section('priority').items():
                mod_priority[name] = int(priority)

        # We will load all recognized modules.
        name_pattern = re.compile(r'([0-9][0-9])_(.*)\.py')
        dirs = []
        for user_dir in self.cfg.mod_dirs():
            if user_dir not in dirs:
                dirs.append(user_dir)
        dirs += [Path(__file__).parent.resolve() / 'modules']
        mods = []
        for mdir in dirs:
            for mod_file in sorted(os.listdir(mdir)):
                m = name_pattern.match(mod_file)
                if m:
                    path = (mdir / mod_file).resolve()
                    name = m.group(2)
                    loader = importlib.machinery.SourceFileLoader(name, str(path))
                    spec = importlib.util.spec_from_loader(name, loader)
                    mod = importlib.util.module_from_spec(spec)
                    loader.exec_module(mod)
                    if name in mod_priority:
                        priority = mod_priority[name]
                    else:
                        priority = int(m.group(1))
                    mods.append((priority, name, mod))

        # Initialize in priority order.
        for _, name, mod in sorted(mods):
            self.print(f'Init:', mod.__doc__ if mod.__doc__ else name)
            mod.init(self)

    def find_entrypoint(self, protocol, hostname, path):
        try:
            for entry in self.entrypoints[protocol][hostname]:
                path_pattern, handler = entry
                if handler != None:
                    # A path string, possibly with wildcards.
                    if len(path_pattern) == 0 or fnmatch.fnmatch(path, path_pattern):
                        return handler
                else:
                    # A callable generic path matcher.
                    handler = path_pattern(path)
                    if handler:
                        return handler
        except Exception as x:
            print(x)
            return None

        return None

    def call_entrypoint(self, request):
        """
        Calls the registered entry point for a request.

        Args:
            request (Request): Request object.

        Returns:
            tuple: (status, meta, body, cache). The type of body can be bytes,
            bytearray, or pathlib.Path. Returning a pathlib.Path means that
            the body will be read from the referenced file. The cache is None
            if the data was not read from a cache.
        """
        entrypoint = self.find_entrypoint(request.scheme, request.hostname, request.path)

        caches = self.caches if (request.scheme == 'gemini' and
                                 not request.identity and
                                 not request.query) else []
        from_cache = None

        if entrypoint:
            # Check the caches first.
            for cache in caches:
                media, content = cache.try_load(request.hostname + request.path)
                if not media is None:
                    if hasattr(content, '__len__'):
                        self.print('%d bytes from cache, %s' % (len(content), media))
                    else:
                        self.print('stream from cache,', media)
                    return 20, media, content, cache

            # Process the request normally if there is nothing cached.
            if not from_cache:
                try:
                    if not self.handler_queue:
                        # Handle in the same thread/process synchronously. This is probably
                        # running under a RequestHandler process.
                        response = entrypoint(request)
                    else:
                        # Put it in the handler queue and wait for completion. Parser threads use
                        # this to hand work off to the handler processes.
                        with self.job_lock:
                            # The job ID is for verifying we are getting the right response.
                            self.job_id += 1
                            job_id = self.job_id

                        self.handler_queue.put((job_id, request, request.worker_id))
                        result_id, response = self.response_queues[request.worker_id].get()

                        if result_id != job_id:
                            raise Exception('response queue out of sync: request handler returned wrong job ID')
                        if isinstance(response, Exception):
                            raise response

                    status, meta, body = unpack_response(response)

                    return status, meta, body, None

                except Exception as x:
                    import traceback
                    traceback.print_exception(x)
                    raise GeminiError(40, 'Temporary failure')

        raise GeminiError(50, 'Permanent failure')


class RequestData:
    """Encapsules data about an incoming request, before it has been fully parsed."""

    def __init__(self, worker, stream, buffered_data, from_addr, identity, request):
        self.worker = worker
        self.stream = stream
        self.buffered_data = buffered_data
        self.from_addr = from_addr
        self.identity = identity
        self.request = request


class RequestParser(threading.Thread):
    """Thread that parses incoming requests from clients."""

    def __init__(self, id, context, job_queue):
        super().__init__()
        self.id = id
        self.context = context
        self.cfg = context.cfg
        self.port = self.cfg.port()
        self.jobs = job_queue

    def run(self):
        try:
            while True:
                stream, from_addr = self.jobs.get()
                if stream is None:
                    break
                try:
                    self.process_request(stream, from_addr)
                except OpenSSL.SSL.SysCallError as error:
                    self.log(f'OpenSSL error: ' + str(error))
                except AbortedIOError as error:
                    self.log(f'Send aborted: ' + str(error))
                except Exception as error:
                    self.log(f'Problem: ' + str(error))
                    # Some unexpected problem...
                    #import traceback
                    #traceback.print_exc()
                    # try:
                    #     report_error(stream, 42, str(error))
                    # except:
                    #     pass
                safe_close(stream)
                stream, from_addr = None, None
        except KeyboardInterrupt:
            pass

    def log(self, *args):
        print(time.strftime('%Y-%m-%d %H:%M:%S'), f'[{self.id}]', '--', *args)

    def process_request(self, stream, from_addr):
        data = bytes()
        MAX_LEN = 1024  # TODO: Gemini/Titan limitation only.
        MAX_RECV = MAX_LEN + 2  # includes terminator "\r\n"
        request = None
        incoming = safe_recv(stream, MAX_RECV)

        try:
            while len(data) < MAX_RECV:
                data += incoming
                crlf_pos = data.find(b'\r\n')
                if crlf_pos >= 0:
                    request = data[:crlf_pos].decode('utf-8')
                    data = data[crlf_pos + 2:]
                    break
                elif len(data) > MAX_LEN:
                    # At this point we should have received the line terminator.
                    self.log(from_addr, 'sent a malformed request')
                    report_error(stream, 59, "Request exceeds maximum length")
                    return

                incoming = safe_recv(stream, MAX_RECV - len(data))
                if len(incoming) <= 0:
                    break
        except UnicodeDecodeError:
            report_error(stream, 59, "Request contains malformed UTF-8")
            return

        if not request or len(data) > MAX_RECV:
            report_error(stream, 59, "Bad request")
            return

        cl_cert = stream.get_peer_certificate()
        identity = Identity(cl_cert) if cl_cert else None

        for scheme, handler in self.context.protocols.items():
            if request.startswith(scheme + ':'):
                self.log(request)
                response = handler(RequestData(self, stream, data, from_addr, identity, request))
                if not response is None:
                    safe_sendall(stream, response)
                return

        report_error(stream, 59, "Unsupported protocol")


class ServerRestarter:
    def __init__(self, server):
        self.server = server

    def __call__(self, signum, frame):
        if signum == signal.SIGHUP:
            print('--- SIGHUP ---')
            self.server.restart_workers()


class RequestHandler(mp.Process):
    def __init__(self, id, cfg, job_queue, result_queues):
        super().__init__(target=RequestHandler._run, args=(self,))
        self.id = id
        self.cfg = cfg
        self.jobs = job_queue
        self.results = result_queues
        self.context = None

    def _run(self):
        self.context = Context(self.cfg)
        self.context.load_modules()

        # Wait for request processing jobs.
        try:
            while True:
                job_id, request, queue_id = self.jobs.get()
                if job_id is None:
                    break
                result_queue = self.results[queue_id]
                entrypoint = self.context.find_entrypoint(request.scheme, request.hostname, request.path)
                if not entrypoint:
                    result_queue.put((job_id, Exception("Missing entrypoint: " + request.url())))
                    continue
                try:
                    response = unpack_response(entrypoint(request))
                    result_queue.put((job_id, response))

                except Exception as error:
                    result_queue.put((job_id, error))

        except KeyboardInterrupt:
            pass


class Server:
    def __init__(self, cfg):
        mp.set_start_method('spawn')

        self.cfg = cfg
        self.address = cfg.address()
        self.port = cfg.port()

        cert_path = cfg.certs_dir() / 'cert.pem'
        key_path = cfg.certs_dir() / 'key.pem'

        if not os.path.exists(cert_path):
            raise Exception("certificate file not found: " + str(cert_path))
        if not os.path.exists(key_path):
            raise Exception("private key file not found: " + str(key_path))

        self.context = SSL.Context(SSL.TLS_SERVER_METHOD)
        self.context.use_certificate_file(str(cert_path))
        self.context.use_privatekey_file(str(key_path))
        self.context.set_verify(SSL.VERIFY_PEER, verify_callback)

        session_id = f'GmCapsule:{cfg.port()}'.encode('utf-8')
        #if type(session_id) != bytes:
        #    raise Exception("session_id type must be `bytes`")
        self.context.set_session_id(session_id)

        # Spawn the worker threads.
        self.parser_queue = queue.Queue()
        self.handler_queue = mp.Queue()
        self.init_parser_context()
        self.parsers = []
        self.handlers = []
        self.create_workers(cfg)

        self.sock = None
        self.sv_conn = None

    def run(self):
        attempts = 60
        print(f'Opening port {self.port}...')
        while True:
            try:
                self.sock = socket.socket()
                self.sock.bind((self.address, self.port))
                self.sock.listen(5)
                self.sv_conn = SSL.Connection(self.context, self.sock)
                self.sv_conn.set_accept_state()
                break
            except:
                attempts -= 1
                if attempts == 0:
                    raise Exception(f'Failed to open port {self.port} for listening')
                time.sleep(2.0)
                print('...')
        print(f'Server started on port {self.port}')

        self.start_workers()

        try:
            signal.signal(signal.SIGHUP, ServerRestarter(self))
        except ValueError:
            print('Restarting with SIGHUP not supported')

        while True:
            stream = None
            try:
                stream, from_addr = self.sv_conn.accept()
                stream._socket.settimeout(10)
                self.parser_queue.put((stream, from_addr))
                del stream
                del from_addr
            except KeyboardInterrupt:
                print('\nStopping the server...')
                break
            except Exception as ex:
                #import traceback
                #traceback.print_exc()
                print(ex)

        # Close the server socket.
        self.sv_conn = None
        self.sock.close()
        self.sock = None

        # Stop all workers.
        self.stop_workers()

        print('Done')

    def init_parser_context(self):
        self.handler_results = []
        if self.is_using_handler_processes():
            for _ in range(max(self.cfg.num_threads(), 1)):
                # Handler processes put results in these queues.
                self.handler_results.append(mp.Queue())
        self.parser_context = Context(self.cfg,
                                      allow_extension_workers=True,
                                      handler_queue=self.handler_queue if self.is_using_handler_processes() else None,
                                      response_queues=self.handler_results if self.is_using_handler_processes() else None)
        self.parser_context.set_quiet(False)
        self.parser_context.load_modules()

    def restart_workers(self):
        """
        Restarts workers with an updated configuration. The server socket or
        TLS configuration are not modified, even if the values have changed
        in the configuration file.
        """
        self.stop_workers()
        self.cfg.reload()
        self.init_parser_context()
        self.create_workers(self.cfg)
        self.start_workers()

    def is_using_handler_processes(self):
        return self.cfg.num_processes() > 0

    def create_workers(self, cfg):
        for proc_id in range(max(cfg.num_processes(), 0)):
            proc = RequestHandler(proc_id, cfg, self.handler_queue, self.handler_results)
            self.handlers.append(proc)

        for parser_id in range(max(cfg.num_threads(), 1)):
            # Threads share one context (note: GIL).
            parser = RequestParser(parser_id, self.parser_context, self.parser_queue)
            self.parsers.append(parser)

    def start_workers(self):
        for handler in self.handlers:
            handler.start()
        for parser in self.parsers:
            parser.start()

        print(len(self.parsers), 'parser(s) and', len(self.handlers), 'handler(s) started')

    def stop_workers(self):
        self.parser_context.shutdown.set()

        # Stop parsers first so all ongoing handler processes get to finish, and no new
        # requests can come in.
        for _ in range(len(self.parsers)):
            self.parser_queue.put((None, None))
        for _ in range(len(self.handlers)):
            self.handler_queue.put((None, None, None))

        for p in self.parsers:
            p.join()
        for h in self.handlers:
            h.join()

        print(len(self.parsers), 'parser(s) and', len(self.handlers), 'handler(s) stopped')
        self.parsers = []
        self.handlers = []
        self.parser_context.shutdown.clear()



# Copyright (c) 2022 Jaakko Keränen <jaakko.keranen@iki.fi>
# License: BSD-2-Clause

"""

User manual
===========

GmCapsule is an extensible Gemini/Titan server.

Extensibility is achieved with Python modules that get loaded at launch
from the configured directories. A set of built-in extension modules is
provided for common functionality like CGI and for serving static files.

The supported protocols are `Gemini <https://gemini.circumlunar.space>`_ and
`Titan <https://transjovian.org/titan>`_. Both are accepted via the same
TCP port.

GmCapsule can be used in a few different ways:

- You can run it as-is for serving static files.
- You can use CGI programs to generate dynamic content and/or process
  queries and uploads. As an extreme example, you could attach a CGI
  program to the path ``/*`` and generate the entire capsule procedurally
  with it.
- You can use the included extension module `gitview` to make local Git
  repositories viewable via Gemini.
- You can write new extension modules that run as part of the server
  process for advanced use cases. For example, this enables the use
  of additional worker threads and caching state in memory.

``gmcapsuled`` is a simple command line utility for loading a configuration
and running the server. Use the ``--help`` option to see usage instructions.

One can also run the server manually in Python::

    from gmcapsule import *
    cfg = Config("myconfig.ini")
    capsule = Capsule(cfg)
    capsule.run()


Getting started
***************

1. Acquire or generate a server certificate.
2. Prepare a configuration file. A configuration file is required for
   running GmCapsule. For more details, see :ref:`Configuration` and
   :ref:`example.ini`.
3. Run ``gmcapsuled``.


Configuration
*************

The GmCapsule configuration file is in `INI format
<https://en.wikipedia.org/wiki/INI_file>`_. The following sections are
defined:

- :ref:`server` — server settings
- :ref:`titan` — Titan upload settings
- :ref:`static` — serving static files
- :ref:`rewrite.*` — URL rewriting rules
- :ref:`cgi` — General CGI settings
- :ref:`cgi.*` — CGI programs
- :ref:`gitview` — Git repository viewer settings
- :ref:`gitview.*` — Git repository settings

Example of a minimal configuration file for serving static files from
`/var/gemini/example.com/`:

.. code-block:: ini

    [server]
    host  = example.com
    certs = /home/username/.certs

    [static]
    root = /var/gemini


server
------

host : string [string...]
    One or more hostnames for the server. Defaults to ``localhost``.

address : string
    IP address of the network interface where the server is listening.
    Defaults to ``0.0.0.0`` (all interfaces).

port : int
    IP port on which the server is listening.

certs : path
    Directory where the server certificate is stored. The directory must
    contain the PEM-formatted files `cert.pem` and `key.pem`. Defaults
    to `.certs`.

modules : path [path...]
    One or more directories to load extension modules from.

threads : int
    Number of worker threads. At least 1 is required. Defaults to 5.
    Determines how many clients can be responded to at the same time.

processes : int
    Number of request handler processes. Defaults to 2. Request handler
    processes run independently of the main process, allowing multiple
    long-running requests to be handled concurrently. If zero, the worker
    threads handle requests synchronously and are able to only do I/O
    operations concurrently.


titan
-----

Settings for Titan requests.

upload_limit : int
    Maximum size of content accepted in an upload, in bytes. Defaults to
    ``10485760`` (i.e., 10 MiB). Requests that attempt to upload larger
    content will be rejected with an error status.


static
------

Settings for the `static` module that serves files from a directory. Omitting
this section from the configuration disables the module.

root : path [path...]
    Content directory. Defaults to `.` (current directory). The hostname
    is appended as a subdirectory, so for example if this is set to:

    .. code-block:: ini

        [static]
        root = /home/user/gemini

    files will be served from `/home/user/gemini/example.com/`.


rewrite.*
---------

Settings for the `rewrite` module that checks regular expressions against
the request path and can rewrite the path or return a custom status. You can
use this for internal remapping of directories and files, redirections,
"Gone" statuses, or other exceptional situations.

Each rewriting rule is a section that begins with ``rewrite.``.

.. code-block:: ini

    [rewrite.rename]
    path    = ^/old-path/
    repl    = /new-location/

    [rewrite.elsewhere]
    path    = .*\\.gmi$
    status  = 31 gemini://mygemlog.space/\\1.gmi

protocol : string
    Protocol for the rewrite rule. If omitted, the rule applies to both
    ``gemini`` and ``titan``.

host : string
    Hostname for the rewrite rule. If omitted, defaults to the first
    hostname defined in the :ref:`server` section.

path : string
    Regular expression that is matched against the request path. You may use
    capture groups and refer to them in the replacement text. Note that the
    request path always begins with a slash.

repl : string
    Replacement path. The part of the request path that matches the "path"
    pattern is replaced with this. You can use backslashes to refer to
    capture groups (``\\1``).

status : string
    Custom status to respond with. Must begin with the status code followed
    by the meta line. You can use backslashes to refer to capture groups
    (``\\1``).


cgi
---

General settings for CGI programs.

bin_root : path
    CGI root directory. If set, all executables under the root are made
    available at corresponding URL entry points. The entire directory tree is
    checked for executables. This mapping of executables to entry points is
    dynamic, so you can add, modify, and remove executables inside the
    directory tree without restarting the server.

    The hostname is appended as a subdirectory, like with the static root
    content directory. For example:

    .. code-block:: ini

        [cgi]
        bin_root = /home/user/gemini/cgi-bin

    An executable at `/home/user/gemini/cgi-bin/example.com/action` would then
    be visible at gemini://example.com/action. If the executable name ends
    with ``,titan`` (note: a comma), the entrypoint will use the Titan
    protocol instead of Gemini. The ``,titan`` suffix is omitted from the URL.


cgi.*
-----

Each section whose name begins with ``cgi.`` is used for setting up CGI entry
points for the `cgi` module. For example, this registers
``gemini://example.com/listing`` to show the output of ``/bin/ls -l``:

.. code-block:: ini

    [cgi.listfiles]
    path    = /listing
    cwd     = /home/username/testdir
    command = /bin/ls -l

protocol : string
    Protocol for the CGI entry point. Defaults to ``gemini``.

host : string
    Hostname for the CGI entry point. If omitted, defaults to the first
    hostname defined in the :ref:`server` section.

path : string [string...]
    URL path for CGI entry point. An asterisk ``*`` in the end is considered
    a wildcard, matching any path that begins with the specified path.
    Multiple paths can be specified.

cwd : path
    Working directory for executing the CGI command. If omitted, the CGI
    command is executed in the directory where the server was started from.

command : string [string...]
    Command to execute when a request is received matching the CGI entry
    point. The specified strings are interpreted like a shell command, i.e.,
    values that contain spaces need to be quoted.

    See :ref:`CGI Programs` for details about executing CGI programs.


gitview
-------

Settings for the `gitview` module that enables viewing local Git
repositories via Gemini. If this section is missing, `gitview` is disabled.

host : string
    Hostname where is `gitview` running. If omitted, defaults to the first
    hostname defined in the :ref:`server` section. `gitview` is limited to a
    single hostname.

    This version of `gitview` assumes that the host is reserved just
    for viewing Git repositories. Therefore, using a dedicated virtual host
    is recommended, for example `git.example.com`. The hostname must be
    one of the hostnames in the :ref:`server` section.

git : path
    Path of the Git executable to use. Defaults to `/usr/bin/git`.

cache_path : path
    Directory where cached pages are stored. If omitted, caching is disabled.


gitview.*
---------

Configuration of a Git repository that is visible via `gitview`. For example:

.. code-block:: ini

    [gitview.lagrange]
    title          = Lagrange
    brief          = A Beautiful Gemini Client
    clone_url      = https://git.skyjake.fi/gemini/lagrange.git
    tag_url        = https://git.skyjake.fi/gemini/lagrange/releases/tag/{tag}
    path           = /Users/jaakko/src/lagrange
    url_root       = lagrange
    default_branch = release

title : string
    Name of the repository.

brief : string
    Short description of the repository. Shown on the root page.

clone_url : url
    Git clone URL.

tag_url : url
    Git tag URL for viewing tags in a web browser. This is useful for showing
    release notes and downloads attached to the tag. The placeholder ``{tag}``
    is replaced with the tag to view.

path : path
    Local file path of the Git repository. Git will be run in this directory.

url_root : string
    URL root path where the Git repository can be viewed via Gemini.

default_branch : string
    The default branch of the repository. This is used if no branch is
    specified in the URL.


Static files
************

Static files are served from the configured content directory. There must be
a separate subdirectory for each configured hostname inside the content
directory. For example, if the content directory is `/var/gemini`::

    var
    └── gemini
        ├── example.com
        └── sub.example.com

If the requested path is a directory that contains an `index.gmi` file, that
file is automatically sent as a response.

If the requested path is a directory but the URL path does not end with
a slash, a redirect is sent as a response with the slash added.

The MIME types of files are autodetected using Python's ``mimetypes`` and the
system ``file`` utility, if that is available. A few common file extensions
like ``.txt`` are directly mapped to the corresponding MIME types.


.meta files
-----------

A `.meta` file can be placed in any directory inside the content directory
tree, containing additional directives for handling static files.

Each `.meta` file affects the directory it's in and also all its
subdirectories. For example, placing a `.meta` in `/var/gemini` would affect
all directories inside all virtual hosts.

The `.meta` file has a simple format. Each line has the following structure:

.. code-block:: rst

    pattern: directive

``pattern``
    Shell globbing expression. For example: ``*.zip``
``directive``
    MIME type, or the full Gemini header that will be sent when serving
    the file.

If only a MIME type is specified, that will be used instead of the
autodetected type. If the full header is specified with a non-20 status code,
the response body will be empty and only the header is sent.


CGI programs
************

CGI programs inherit the environment of the server process. Additionally, the
following CGI-specific environment variables are set, providing input and
contextual information:

- ``REMOTE_ADDR``
- ``QUERY_STRING``
- ``PATH_INFO``
- ``SCRIPT_NAME``
- ``SERVER_PROTOCOL``
- ``SERVER_NAME``
- ``SERVER_PORT``
- ``AUTH_TYPE``
- ``TLS_CLIENT_HASH`` (when client certificate provided)
- ``REMOTE_IDENT`` (fingerprints of client certificate and public key)
- ``REMOTE_USER`` (when client certificate provided)
- ``CONTENT_LENGTH`` (Titan only)
- ``CONTENT_TYPE`` (Titan only)
- ``TITAN_TOKEN`` (Titan only)

The CGI programs's stdout is used as the response to the request. The response
is expected to begin with a valid meta line (status code and meta text), but
this is not required. If the meta line is missing, the following header is
prepended to UTF-8 text output:

.. code-block:: rst

    20 text/plain; charset=utf-8

Or, if the output is binary, the prepended header will be:

.. code-block:: rst

    20 application/octet-stream

The response is sent to the client after the program finishes running.

With Titan requests, the uploaded content is passed in via stdin. Note that
the CGI program is only executed after the full contents have already been
successfully received, so the program does not need to worry about interrupted
uploads.


Extensions
**********

All Python modules in the configured extension module directories (see
:ref:`server` configuration) are loaded at launch in alphabetical order, as
long as they use the following naming convention:

    `NN_extmod.py`

That is, the name of the extension ("extmod") is prefixed with two numbers
`NN` and an underscore. This naming gives more control over the order in
which modules are loaded. The ones loaded first have precedence over
registered URL entry points.

If the server has been configured to use multiple processes, each process
loads the extensions separately. This may require synchronizing access to
external resources accessed by extensions.


Initialization
--------------

Each extension module is required to have an initialization function:

.. py:function:: extmod.init(context)

    Initialize the extension module.

    This is called once immediately after the extension has been loaded.

    :param context: Worker context. The extension can use this to access
        configuration parameters, install caches, and register entry
        points and custom scheme handlers.
    :type context: gmcapsule.Context


Requests
--------

An extension module is responsible for handling requests arriving at its
registered entry points. A :class:`~gmcapsule.gemini.Request` object is
given as an argument when the registered entry point callback or handler
object gets called.

The return value from the handler determines what gets sent as a response
to the client:

- Returning ``bytes`` or ``str`` means that the response is UTF-8 encoded
  "text/gemini", and the status code is 20.
- Returning a two-element tuple is interpreted as (status code, meta).
  The response body will be empty. This is useful for error messages and
  redirects.
- Returning a three-element tuple is interpreted as (status, meta, body).
  If the body type is ``str``, it will be UTF-8 encoded before sending.
  If the body type is ``pathlib.Path``, the referenced file is opened and
  its contents are streamed to the client without reading the entire file
  to memory. Otherwise, the body type must be ``bytes`` or ``bytearray``.


Future improvements
*******************

The following limitations could/should be lifted in the future:

- Enable value interpolation in the `configparser`, allowing access to
  defined values and environment variables.
- `gitview` should have a URL path prefix.
- Caches should be informed of identities and queries; content might still
  be cacheable.

"""

import configparser
import importlib
import importlib.machinery
import importlib.util
import mimetypes
import os
import re
import shlex
import subprocess
from pathlib import Path

from .gemini import Server, Cache, Context
from .markdown import to_gemtext as markdown_to_gemtext


__version__ = '0.5.0'
__all__ = [
    'Config', 'Cache', 'Context',
    'get_mime_type', 'markdown_to_gemtext'
]


class Config:
    """
    Server configuration.

    Args:
        config_path (str): Path of a config INI file to load.

    Attributes:
        ini(configparser.ConfigParser): Contents of the config INI file.
            Extension modules can use this to access additional custom
            config parameters. See `configparser
            <https://docs.python.org/3/library/configparser.html>`_
            for details.
    """

    def __init__(self, config_path):
        self.config_path = config_path
        self.reload()

    def reload(self):
        self.ini = configparser.ConfigParser()
        if os.path.exists(self.config_path):
            print('Configuration:', self.config_path)
            self.ini.read(self.config_path)
        else:
            print(self.config_path, 'not found -- using defaults')

    def hostnames(self):
        """
        Returns:
            list(str): All the configured hostnames. The first listed hostname
            is considered the default to be used when a hostname is not
            otherwise specified.
        """
        return self.ini.get('server', 'host', fallback='localhost').split()

    def address(self):
        return self.ini.get('server', 'address', fallback='0.0.0.0')

    def port(self):
        """
        Returns:
            int: Listening IP port of the server.
        """
        return self.ini.getint('server', 'port', fallback=1965)

    def certs_dir(self):
        return Path(self.ini.get('server', 'certs', fallback='.certs'))

    def root_dir(self):
        """
        Returns:
            pathlib.Path: Content root directory for serving static files.
            The hostname is always automatically appended to this as a
            subdirectory.
        """
        return Path(self.ini.get('static', 'root')).resolve()

    def mod_dirs(self):
        return [Path(p).resolve() for p in shlex.split(
            self.ini.get('server', 'modules', fallback='.'))]

    def num_threads(self):
        return self.ini.getint('server', 'threads', fallback=5)

    def num_processes(self):
        return self.ini.getint('server', 'processes', fallback=2)

    def max_upload_size(self):
        return self.ini.getint('titan', 'upload_limit', fallback=10 * 1024 * 1024)

    def require_upload_identity(self):
        return self.ini.getboolean('titan', 'require_identity', fallback=True)

    def section(self, name):
        """
        Find a section in the config INI file.

        Args:
            name (str): Name of the section.

        Returns:
            configparser.SectionProxy: INI section.

        Raises:
            KeyError: The section was not found.
        """
        return self.ini[name]

    def prefixed_sections(self, prefix):
        """
        Find all sections in the config INI file whose name begins with
        the given prefix.

        Args:
            prefix (str): Name prefix, e.g., ``cgi.``.

        Returns:
            dict: Mapping of section names (with the prefix removed) to the
            corresponding INI sections (configparser.SectionProxy). An
            empty dictionary is returned if there are no sections matching
            the prefix.
        """
        sects = {}
        for name in self.ini.sections():
            if not name.startswith(prefix): continue
            sects[name[len(prefix):]] = self.ini[name]
        return sects


class Capsule:
    """
    Server instance.

    The server is initialized as specified in the configuration.
    Extension modules are loaded and initialized.

    After constructing and setting up Capsule, call the
    :func:`~gmcapsule.Capsule.run` method to begin accepting incoming
    connections.

    Args:
        cfg (Config): Server configuration.
    """

    def __init__(self, cfg):
        self.cfg = cfg
        self.sv = Server(cfg)

    def run(self):
        """
        Start worker threads and begin accepting incoming connections. The
        server will run until stopped with a KeyboardInterrupt (^C).
        """
        self.sv.run()


def get_mime_type(path):
    """
    Determine the MIME type of a file. A handful of common file extensions
    are detected as special cases, such as ".gmi" and ".txt". Other files
    are detected with Python's ``mimetypes`` standard library module, and as
    a final fallback, the ``file`` command line utility.

    Args:
        path (str): File path.

    Returns:
        str: Detected MIME type, for example, "image/jpeg". Returns
        "application/octet-stream" if the correct type could not be
        determined.
    """
    p = str(path)
    lp = p.lower()
    if lp.endswith('.txt'):
        return 'text/plain'
    if lp.endswith('.gmi') or lp.endswith('.gemini'):
        return 'text/gemini'
    if lp.endswith('.md') or lp.endswith('.markdown') or lp.endswith('.mdown'):
        return 'text/markdown'

    if not Path(p).exists():
        return None

    mt = mimetypes.guess_type(p)[0]
    if mt is not None:
       return mt

    try:
        return subprocess.check_output([
            '/usr/bin/env', 'file', '--mime-type', '-b', p
            ]).decode('utf-8').strip()
    except:
        return 'application/octet-stream'


def run_server():
    print(f"GmCapsule v{__version__}")

    import argparse
    argp = argparse.ArgumentParser(
        description='GmCapsule is an extensible server for Gemini and Titan.')
    argp.add_argument('-c', '--config',
                      dest='config_file',
                      default=Path.home() / '.gmcapsulerc',
                      help='Configuration file to load at startup')
    argp.add_argument('--trace-malloc',
                      action='store_true',
                      help='Enable memory allocation tracing (for debugging)')
    args = argp.parse_args()

    cfg = Config(args.config_file)
    try:
        capsule = Capsule(cfg)
        capsule.run()
        return 0
    except Exception as er:
        print('ERROR:', er)
        return -1



# Copyright (c) 2023 Jaakko Keränen <jaakko.keranen@iki.fi>
# License: BSD-2-Clause

"""Rewriter"""

import re


class PathRewriteHandler:
    def __init__(self, context, rewritten_path):
        self.context = context
        self.rewritten_path = rewritten_path

    def __call__(self, req):
        old_path = req.path
        req.path = self.rewritten_path

        # Don't allow rewriting the same request too many times.
        if hasattr(req, 'num_rewrites'):
            req.num_rewrites += 1
        else:
            req.num_rewrites = 1
        if req.num_rewrites == 100:
            return 40, "Stuck in rewrite loop: " + req.url()

        self.context.print("[rewrite]", old_path, "->", req.path)
        return self.context.call_entrypoint(req)[0]


class Responder:
    def __init__(self, code, meta):
        self.code = code
        self.meta = meta

    def __call__(self, req):
        meta = self.meta.replace('${QUERY_STRING}', f'?{req.query}' if req.query != None else '')
        return self.code, meta


class Rewriter:
    def __init__(self, context, protocol, host, src_path, dst_path, status):
        self.context = context
        self.protocol = protocol
        self.host = host
        self.src_path = src_path
        self.dst_path = dst_path
        self.status = status

    def __call__(self, path):
        # If path matches a rewritten URL, return the handler object that calls the
        # correct handler for the updated URL.
        if self.dst_path:
            new_path = self.src_path.sub(self.dst_path, path)
            if new_path != path:
                return PathRewriteHandler(self.context, new_path)

        elif self.status:
            m = self.src_path.match(path)
            if m:
                status = self.status
                for i in range(self.src_path.groups + 1):
                    cap = m[i]
                    if cap:
                        status = status.replace(f'\\{i}', cap)
                code, meta = status.split()
                self.context.print("[rewrite]", code, meta)
                return Responder(int(code), meta)

        return None


def init(context):
    cfg = context.config()
    for section in cfg.prefixed_sections('rewrite.').values():
        protocol = section.get('protocol', None)
        host = section.get('host', cfg.hostnames()[0])
        src_path = re.compile(section.get('path'))
        dst_path = section.get('repl', None)
        status = section.get('status', None)
        for proto in [protocol] if protocol else ['gemini', 'titan']:
            context.add(Rewriter(context, proto, host, src_path, dst_path, status),
                        None, # `Rewriter` will return a suitable handler callback.
                        host,
                        proto)



# Copyright (c) 2021-2022 Jaakko Keränen <jaakko.keranen@iki.fi>
# License: BSD-2-Clause

"""Git Repository Viewer"""

import hashlib
import json
import os
import os.path
import pickle
import re
import subprocess
import time
import urllib
from pathlib import Path

from gmcapsule import Cache, markdown_to_gemtext

pjoin = os.path.join


class GitViewCache(Cache):
    """
    File-based cache that stores content using ``pickle``.

    File paths inside the cache directory are based on a SHA-256 hash of
    the original URL path.

    Args:
        hostname (str): Hostname that this cache belongs to.
        file_root (str): Directory where the cached content is stored.
            A large number of hash-prefix subdirectories are created here.
    """
    def __init__(self, hostname, file_root):
        super().__init__()
        self.hostname = hostname
        self.file_root = file_root
        os.makedirs(file_root, exist_ok=True)
        self.ptn_static_path = re.compile('.*/(tags|commits|patch|cdiff|pcdiff)/[0-9a-f]+')

    def storage_path(self, path):
        digest = hashlib.sha256(path.encode('utf-8')).hexdigest()
        return digest[0:2] + '/' + digest[2:]

    def max_age(self, path):
        if self.ptn_static_path.match(path):
            return 60 * 24 * 3600  # two months
        return 3600 # one hour for dynamic content

    def try_load(self, path):
        if not path.startswith(self.hostname + '/'):
            return None, None
        storage_path = pjoin(self.file_root, self.storage_path(path))
        if not os.path.exists(storage_path):
            return None, None
        now_time = time.time()
        storage_time = os.path.getmtime(storage_path)
        # print('cache {%s} age %d seconds' % (path, int(now_time - storage_time)))
        if now_time - storage_time > self.max_age(path):
            return None, None
        return pickle.load(open(storage_path, 'rb'))

    def save(self, path, media_type, content):
        if not path.startswith(self.hostname + '/'):
            return False
        storage_path = pjoin(self.file_root, self.storage_path(path))
        os.makedirs(os.path.dirname(storage_path), exist_ok=True)
        pickle.dump((media_type, content), open(storage_path, 'wb'))
        return True


CONFIG = None
GIT = '/usr/bin/git'
HOSTNAME = 'localhost'
NUM_COMMITS_FRONT = 8
NUM_COMMITS_PER_PAGE = 25
NUM_COMMITS_PER_TAG = 10

# Link Icons
LI_TAG  = '\U0001f516'
LI_FILE = '\U0001f5ce'


def ujoin(a, b):
    if b.startswith('/'):
        return b
    if not a.endswith('/'):
        return a + '/' + b
    return a + b


def preformat(raw, alt_text=''):
    raw = raw.replace('\n```', '\n ```')
    return f'```{alt_text}\n' + raw + '\n```\n'


def repositories():
    roots = []
    for name, cfg in CONFIG.prefixed_sections('gitview.').items():
        url = cfg['url_root']
        if not url.startswith('/'): url = '/' + url
        if not url.endswith('/'): url += '/'
        roots.append((name, url, cfg))
    return roots


class Request:
    def __init__(self, gemini_request):
        self.req = gemini_request
        self.cfg = None
        # Extract the repository root from the path.
        for sectname, url_root, repo_cfg in repositories():
            if self.req.path == url_root or self.req.path == url_root[:-1] or \
                    self.req.path.startswith(url_root):
                self.cfg = repo_cfg
                self.url_root = url_root
                break
        # Parse the path.
        self.path = self.req.path
        self.branch = self.cfg['default_branch']
        if self.path.startswith(self.url_root):
            self.path = self.path[len(self.url_root):]
            if '/' in self.path[1:]:
                parts = self.path.split('/')
                if len(parts) > 0 and len(parts[0]) > 0:
                    self.branch = parts[0].replace('%2F', '/')
                    self.path = '/'.join(parts[1:])
        self.ubranch = self.branch.replace('/', '%2F') + '/'

        # print('cfg:', self.cfg)
        # print('url_root:', self.url_root)
        # print('path:', self.path)
        # print('branch:', self.branch)
        # print('ubranch:', self.ubranch)

    def git(self, args, as_bytes=False):
        result = subprocess.check_output(
            [GIT, '-C', self.cfg['path']] + args
        )
        if as_bytes: return result
        return result.decode('utf-8').rstrip()

    def git_commits(self, count, skip=0, commit_hash=None):
        try:
            out = self.git([
                'log',
                f'-n{count}',
                f'--skip={skip}',
                "--pretty=format:{^@^hash^@^:^@^%h^@^,^@^fullHash^@^:^@^%H^@^,^@^parent^@^:^@^%p^@^,^@^refs^@^:^@^%D^@^,^@^author^@^:^@^%an^@^,^@^date^@^:^@^%ad^@^,^@^email^@^:^@^%aE^@^,^@^message^@^:^@^%s^@^,^@^body^@^:^@^%b^@^,^@^commitDate^@^:^@^%ai^@^,^@^age^@^:^@^%cr^@^},^@&@^",
                commit_hash if commit_hash else self.branch])
            out = out.replace('^@&@^\n', '').replace(',^@&@^', '') \
                    .replace('\\', '\\\\') \
                    .replace('\n', '\\n') \
                    .replace('\t', '    ') \
                    .replace('"', '\\"') \
                    .replace('^@^', '"') \
                    .replace('\\n#', '\\n') \
                    .replace('"body":"#', '"body":"')
            out = '[' + out + ']'
            #print(out)
            return json.loads(out)
        except:
            return []

    def git_tags(self):
        tags = []
        try:
            for info in self.git(['for-each-ref', '--sort=creatordate',
                    '--format', '%(refname), %(creatordate:short)', 'refs/tags']).split('\n'):
                ref, date = info.split(',')
                date = date.strip()
                tags.append((date, ref[10:]))
        except:
            pass
        return tags

    def git_diff(self, commit_hash, raw=False):
        out = self.git(['log', '--color', '-n1', '-p', '--format=format:', commit_hash])
        size = len(out.encode('utf-8'))
        MAX_DIFF = 100 * 1024
        is_trunc = len(out) > MAX_DIFF
        if is_trunc: out = out[:MAX_DIFF]
        if raw:
            return out
        out = out.replace('\n```', '\n ```')
        return preformat(out, alt_text=f'Diff of changes in commit {commit_hash}') + \
            ("(truncated output; full size was %.2f KB)\n" % (size / 1024) if is_trunc else '')

    def git_summary(self, commit_hash):
        out = self.git(['log', '--stat', '-n1', '--format=format:', commit_hash])
        return preformat(out, alt_text='Summary of changes')

    def git_patch(self, commit_hash):
        out = self.git(['format-patch', '-N', '-n1', '--stdout', commit_hash])
        return out

    def git_raw(self, path):
        try:
            raw = self.git(['show', '--format=raw', self.branch + ':' + path], as_bytes=True)
            try:
                lc = path.lower()
                if lc.endswith('.png'):
                    media = 'image/png'
                elif lc.endswith('.jpg') or lc.endswith('.jpeg'):
                    media = 'image/jpeg'
                elif lc.endswith('.md') or lc.endswith('.mdown') or lc.endswith('.markdown'):
                    raw = markdown_to_gemtext(raw.decode('utf-8')).encode('utf-8')
                    media = 'text/gemini;charset=utf-8'
                elif lc.endswith('.gmi'):
                    media = 'text/gemini;charset=utf-8'
                else:
                    raw.decode('utf-8', 'strict')
                    media = 'text/plain;charset=utf-8'
            except UnicodeDecodeError:
                media = 'application/octet-stream'
            return media, raw
        except Exception as x:
            print(x)
        return None

    def git_tree(self, path):
        if path == '':
            path = '.'  # it's the root
        elif not path.endswith('/'):
            path += '/'
        files = []
        ptn_tree = re.compile(r"([0-9]+)\s+(\w+)\s+([0-9a-f]+)\s+([-0-9]+)\s+(.*)")
        for line in self.git(['ls-tree', '-l', self.branch, path]).split('\n'):
            m = ptn_tree.match(line.strip())
            if m:
                objname = m.group(5)
                objsize = int(m.group(4)) if m.group(4) != '-' else None
                objtype = m.group(2)
                files.append((objname, objsize, objtype))
        return files

    def git_readme(self):
        try:
            md = self.git(['show', self.branch + ':README.md'])
            text = markdown_to_gemtext(md)
            title = self.page_title(omit_default_branch=True).rstrip()
            if text.startswith(title):
                # Avoid the redundant level 1 heading.
                text = '# README.md\n\n' + text[len(title):].lstrip()
            return text
        except Exception as x:
            print(x)
        return ''

    def url(self, path):
        return ujoin(self.url_root, path)

    def commit_link(self, info):
        url = self.url(ujoin(self.ubranch, f"commits/{info['fullHash']}"))
        date = info['commitDate'][:10]
        msg = info['message']
        tags = []
        for ref in info['refs'].split(','):
            ref = ref.strip()
            if ref.startswith('tag:'):
                tags.append(LI_TAG + ' ' + ref[5:])
        tag_str = '' if len(tags) == 0 else ' \u2014 ' + ', '.join(tags)
        return f"=> {url} {date} {msg}{tag_str}\n"

    def page_title(self, subtitle=None, omit_default_branch=False):
        show_branch = not omit_default_branch or self.branch != self.cfg['default_branch']
        sub = f': {subtitle}' if subtitle != None else ''
        return f"# {self.cfg['title']}{sub} {'[' + self.branch + ']' if show_branch else ''}\n\n"


def handle_request(gemini_request):
    req = Request(gemini_request)
    DEFAULT_BRANCH = req.cfg['default_branch']

    page = req.page_title()

    if gemini_request.query == 'branches':
        # Branch selection menu.
        page = f"# {req.cfg['title']}: Branches\n\n"
        page += "Select the active branch.\n"
        for b in sorted(req.git(['branch']).split('\n')):
            b = b.strip()
            if b.startswith('*'):
                b = b[1:]
            b = b.strip()
            page += f"=> {req.url(b.replace('/', '%2F') + '/')} {b}\n"
        page += f"\n=> {req.url(DEFAULT_BRANCH + '/')} Use default ({DEFAULT_BRANCH})\n"

    elif req.path == 'tags' or req.path == 'tags/':
        tags = req.git_tags()
        if len(tags) == 0:
            page += 'No tagged commits.\n'
        else:
            page += '## Latest Tags\n\n'
            for date, tag in list(reversed(tags))[:8]:
                page += f"=> {req.url(req.ubranch + 'tags/' + tag)} {date} {tag}\n"
            page += '\n## All Tags\n\n'
            for tag in sorted([ref for _, ref in tags]):
                page += f"=> {req.url(req.ubranch + 'tags/' + tag)} {LI_TAG} {tag}\n"
        page += f"\n=> {req.url(req.ubranch)} Repository\n"

    elif req.path.startswith('tags/'):
        tag = req.path[5:]
        page += f'## {tag}\n\n'
        if 'tag_url' in req.cfg:
            page += f"=> {req.cfg['tag_url'].replace('{tag}', tag)} Open {tag} in Web Browser\n\n"
        page += "### History\n"
        commits = req.git_commits(NUM_COMMITS_PER_TAG, commit_hash=req.path[5:])
        first = True
        for commit in commits:
            page += req.commit_link(commit)
            if first:
                first = False
                if len(commit['body']) > 0:
                    page += commit['body'] + "\n"
        page += f"\n=> {req.url(req.ubranch + 'tags')} {LI_TAG} Tags\n"
        page += f"=> {req.url(req.ubranch)} Repository\n"

    elif req.path == 'commits' or req.path == 'commits/':
        page += '## Commits\n\n'
        try:
            skip = int(gemini_request.query)
        except:
            skip = 0
        count = NUM_COMMITS_PER_PAGE
        commits = req.git_commits(count + 1, skip)
        if skip > 0:
            prev_skip = max(0, skip - count)
            page += f"=> {req.url(ujoin(req.ubranch, f'commits?{prev_skip}'))} Previous {count} commits\n\n"
        for commit in commits[:count]:
            page += req.commit_link(commit)
        if len(commits) > count:
            page += f"\n=> {req.url(ujoin(req.ubranch, f'commits?{skip + count}'))} Next {count} commits\n"
        page += f"\n=> {req.url(req.ubranch)} Repository\n"

    elif req.path.startswith('commits/'):
        info = req.git_commits(1, commit_hash=req.path[8:])
        if len(info) == 0:
            return 51, "Not Found"
        info = info[0]
        hash = info['hash']
        full_hash = info['fullHash']
        email_subject = urllib.parse.quote(f"{req.cfg['title']} commit {hash}")
        email_body = urllib.parse.quote("=> gemini://%s:%d%scommits/%s" %
            (HOSTNAME,
            CONFIG.port(),
            req.url_root + req.ubranch,
            full_hash)
        )
        tags = []
        for ref in info['refs'].split(','):
            ref = ref.strip()
            if ref.startswith('tag:'):
                tags.append(ref[5:])
        body = info['body'].strip()
        page += f"## {info['message']}\n\n"
        page += f"=> mailto:{info['email']}?subject={email_subject}&body={email_body} {info['author']}\n"
        page += f"{info['date']}\n"
        if len(body):
            page += "\n" + body + "\n"
        page += req.git_summary(full_hash)
        page += f"=> {req.url(req.ubranch + 'cdiff/' + full_hash)} Diff (Colored)\n"
        page += f"=> {req.url(req.ubranch + 'pcdiff/' + full_hash)} Diff (Colored, Plain Text)\n"
        page += f"=> {req.url(ujoin(req.ubranch, 'patch/' + full_hash + '.patch'))} \U0001f528 Patch\n"
        page += "\n"
        for tag in tags:
            page += f"=> {req.url(req.ubranch + 'tags/' + tag)} {LI_TAG} {tag}\n"
        for parent in info['parent'].split():
            page += f"=> {req.url(ujoin(req.ubranch, 'commits/' + parent))} Parent {parent}\n"
        page += f"=> {req.url(req.ubranch)} Repository\n"

    elif req.path.startswith('cdiff/'):
        hash = req.path[6:]
        info = req.git_commits(1, commit_hash=hash)
        if len(info) == 0:
            return 51, "Not Found"
        info = info[0]
        page += f"## {info['message']}\n\n"
        page += f"=> {req.url(req.ubranch + 'commits/' + hash)} {hash}\n"
        page += req.git_diff(hash)
        return page

    elif req.path.startswith('pcdiff/'):
        hash = req.path[7:]
        info = req.git_commits(1, commit_hash=hash)
        if len(info) == 0:
            return 51, "Not Found"
        return 20, "text/plain", req.git_diff(hash, raw=True)

    elif req.path.startswith('patch/'):
        hash = req.path[6:-6]   # .patch file extension
        info = req.git_commits(1, commit_hash=hash)
        if len(info) == 0:
            return 51, "Not Found"
        info = info[0]
        return 20, "text/plain", req.git_patch(hash)

    elif req.path.startswith('-/'):
        page = req.page_title(subtitle='File Tree')
        path = req.path[2:]
        tree = req.git_tree(path)
        # Links to parent directories.
        if path != '':
            page += f"=> {req.url(req.ubranch + '-/')} /\n"
        pdir = os.path.dirname(os.path.dirname(path))
        while pdir != '':
            page += f"=> {req.url(req.ubranch + '-/' + pdir + '/')} {pdir}/\n"
            pdir = os.path.dirname(pdir)
        for fpath, fsize, ftype in tree:
            if ftype == 'tree':
                page += f"=> {req.url(req.ubranch + '-/' + fpath + '/')} {fpath}/\n"
            elif ftype == 'blob':
                if fsize < 1.0e4:
                    size = '{:.2f} KB'.format(fsize / 1.0e3)
                elif fsize < 1.0e5:
                    size = '{:.1f} KB'.format(fsize / 1.0e3)
                elif fsize < 1.0e6:
                    size = '{:.0f} KB'.format(fsize / 1.0e3)
                else:
                    size = '{:.1f} MB'.format(fsize / 1.0e6)
                page += f"=> {req.url(req.ubranch + fpath)} {LI_FILE} {fpath}\u00a0\u00a0\u00a0\u00a0[{size}]\n"
            else:
                page += f"{fpath}\u00a0\u00a0\u00a0\u00a0[{ftype}]\n"
        page += f"\n=> {req.url(req.ubranch)} Repository\n"

    else:
        # Check for a raw file at this path.
        if len(req.path) > 0:
            raw = req.git_raw(req.path)
            if raw is not None:
                return 20, raw[0], raw[1]

        page = req.page_title(omit_default_branch=True)

        if 'brief' in req.cfg:
            page += req.cfg['brief'] + '\n'

        page += f"=> {req.url(req.ubranch + 'tags')} {LI_TAG} Tags\n"
        page += f"=> {req.url(req.ubranch + '-/')} {LI_FILE} File Tree\n"
        page += f"=> {req.url('?branches')} \u2325 Branches [{req.branch}]\n"
        page += f"=> {req.cfg['clone_url']} Clone URL\n"
        page += "\n## Latest Commits\n\n"
        for commit in req.git_commits(NUM_COMMITS_FRONT):
            page += req.commit_link(commit)
        page += f"=> {req.url(ujoin(req.ubranch, 'commits'))} More...\n"
        readme = req.git_readme()
        if readme:
            page += readme

    return page


def redirect_to_default(gemini_request):
    req = Request(gemini_request)
    if gemini_request.query:
        return handle_request(gemini_request)
    return 30, req.url(req.cfg['default_branch'] + '/')


def main_page(req):
    page = "# Git Repositories\n"
    for name, url_root, repo_cfg in sorted(repositories()):
        page += f"\n=> {url_root + repo_cfg['default_branch'].replace('/', '%2F') + '/'} {repo_cfg['title']}\n"
        if 'brief' in repo_cfg:
            page += repo_cfg['brief'] + '\n'
    return page


def init(context):
    cfg = context.config()

    global CONFIG
    CONFIG = cfg

    try:
        mod_cfg = cfg.section('gitview')

        global GIT
        GIT = mod_cfg.get('git', fallback=GIT)

        global HOSTNAME
        HOSTNAME = mod_cfg.get('host', fallback=None)
        if HOSTNAME is None:
            HOSTNAME = cfg.hostnames()[0]

        if 'cache_path' in mod_cfg:
            context.add_cache(GitViewCache(HOSTNAME, mod_cfg['cache_path']))

        for name, url_root, _ in repositories():
            context.print(f'  Adding repository "{name}"...')
            context.add('/', main_page, hostname=HOSTNAME)
            context.add(url_root[:-1], redirect_to_default, hostname=HOSTNAME)
            context.add(url_root + '*', handle_request, hostname=HOSTNAME)

    except KeyError:
        # GitView not configured.
        pass



# Copyright (c) 2022 Jaakko Keränen <jaakko.keranen@iki.fi>
# License: BSD-2-Clause

"""CGI programs"""

import os
import pathlib
import shlex
import subprocess
import urllib.parse

import gmcapsule


class CgiContext:
    def __init__(self, port, url_path, args, work_dir=None):
        self.port = port
        self.args = args
        self.base_path = url_path
        if self.base_path.endswith('/*'):
            self.base_path = self.base_path[:-2]
        self.work_dir = work_dir

    def __call__(self, req):
        try:
            env_vars = dict(os.environ)

            # Standard CGI variables.
            env_vars['GATEWAY_INTERFACE'] = 'CGI/1.1'
            env_vars['REMOTE_ADDR'] = '%s:%d' % req.remote_address
            if req.query != None:
                env_vars['QUERY_STRING'] = req.query
            env_vars['PATH_INFO'] = req.path
            env_vars['SCRIPT_NAME'] = self.base_path
            env_vars['SERVER_SOFTWARE'] = 'GmCapsule/' + gmcapsule.__version__
            env_vars['SERVER_PROTOCOL'] = req.scheme.upper()
            env_vars['SERVER_NAME'] = req.hostname
            env_vars['SERVER_PORT'] = str(self.port)
            env_vars[req.scheme.upper() + '_URL'] = f"{req.scheme}://{req.hostname}{req.path}" + (
                '?' + req.query if req.query != None else '')
            env_vars[req.scheme.upper() + '_URL_PATH'] = req.path

            # TLS client certificate.
            if req.identity:
                env_vars['AUTH_TYPE'] = 'Certificate'
                id_sub = req.identity.subject()
                env_vars['REMOTE_USER'] = id_sub['CN'] if 'CN' in id_sub else ''
                env_vars['TLS_CLIENT_HASH'] = req.identity.fp_cert
                env_vars['TLS_CLIENT_SUBJECT'] = ''.join([f'/{k}={v}' for k, v in id_sub.items()])  # "/CN=name"
                env_vars['TLS_CLIENT_ISSUER'] = ''.join([f'/{k}={v}' for k, v in req.identity.issuer().items()])  # "/CN=name"
                env_vars['REMOTE_IDENT'] = str(req.identity)      # cert fingerprints
            else:
                env_vars['AUTH_TYPE'] = ''
                env_vars['REMOTE_USER'] = ''

            # Titan metadata.
            if req.content:
                env_vars['CONTENT_LENGTH'] = str(len(req.content))
                env_vars['CONTENT_TYPE'] = req.content_mime if req.content_mime is not None else ''
                env_vars['TITAN_TOKEN'] = req.content_token if req.content_token is not None else ''
            else:
                env_vars['CONTENT_LENGTH'] = '0'

            result = subprocess.run(self.args,
                                    cwd=self.work_dir,
                                    check=True,
                                    input=req.content,
                                    stdout=subprocess.PIPE,
                                    env=env_vars).stdout
            try:
                # Parse response header.
                crlf_pos = result.find(b'\r\n')
                if crlf_pos >= 1024:
                    return 42, "CGI command returned invalid response header"
                header = result[:crlf_pos].decode('utf-8')
                body = result[crlf_pos + 2:]
                status = int(header[:2])
                meta = header[2:].strip()
                if status < 10:
                    return 42, "CGI command returned invalid status code"
                return status, meta, body
            except:
                try:
                    return 20, 'text/plain; charset=utf-8', result.decode('utf-8')
                except:
                    return 20, 'application/octet-stream', result
        except Exception as er:
            return 42, "CGI error: " + str(er)


class CgiTreeMapper:
    def __init__(self, protocol, host, port, root_dir):
        self.protocol = protocol
        self.host = host
        self.port = port
        self.root_dir = pathlib.Path(root_dir)

    def __call__(self, url_path):
        # Check if url_path is a valid CGI entrypoint and return
        # a CgiContext for it.
        fn = str(self.root_dir) + url_path
        if self.protocol == 'titan':
            fn += ',titan'
        if os.path.isdir(fn):
            return None
        if os.access(fn, os.X_OK):
            return CgiContext(self.port, url_path, [fn], work_dir=os.path.dirname(fn))
        return None


# # NOTE: This require restarting the server when binaries are added/removed.
# def add_cgibin_entrypoints_recursively(context, host, base, cur_dir=None):
#     if cur_dir is None:
#         cur_dir = base
#     for name in os.listdir(cur_dir):
#         fn = cur_dir / name
#         if os.path.isdir(fn):
#             add_cgibin_entrypoints_recursively(context, host, base, fn)
#         elif os.access(fn, os.X_OK):
#             protocol = 'gemini'
#             if name.endswith(',titan'):
#                 protocol = 'titan'
#             # Remove the base directory from the entry path.
#             path = str(fn)[len(str(base)):]
#             args = [str(fn)]
#             if protocol == 'titan':
#                 path = path[:-6]
#             print(f'  {protocol}://{host}{path} ->', args)
#             context.add(path, CgiContext(path, args, work_dir=cur_dir), host, protocol)


def init(context):
    cfg = context.config()
    default_host = cfg.hostnames()[0]

    # Custom entrypoints for specific URLs.
    for section in cfg.prefixed_sections('cgi.').values():
        protocol = section.get('protocol', fallback='gemini')
        host = section.get('host', fallback=default_host)
        work_dir = section.get('cwd', fallback=None)
        args = shlex.split(section.get('command'))
        for path in shlex.split(section.get('path', fallback='/*')):
            context.print(f'  {protocol}://{host}{path} ->', args)
            context.add(path, CgiContext(cfg.port(), path, args, work_dir),
                        host, protocol)

    # Automatic entrypoints for all executables.
    bin_root = cfg.ini.get('cgi', 'bin_root', fallback=None)
    if bin_root != None:
        bin_root = pathlib.Path(bin_root).resolve()
        for host in cfg.hostnames():
            host_bin_root = bin_root / host
            for protocol in ['gemini', 'titan']:
                context.add(
                    CgiTreeMapper(protocol, host, cfg.port(), host_bin_root), None,
                    host, protocol)



# Copyright (c) 2021 Jaakko Keränen <jaakko.keranen@iki.fi>
# License: BSD-2-Clause

import re, string


def sub_cb(pattern, repl, text, match_callback):
    """Substitute all matches of `pattern` with `repl` in `text`,
    calling `match_callback` for each matched region."""
    for m in pattern.finditer(text):
        match_callback(m)
    return pattern.subn(repl, text)[0]


def to_gemtext(src):
    src = src.replace('&nbsp;', '\u00a0')
    src = src.replace('```', '\n```\n')

    ptn_standalone_link = re.compile("^[\\s*_]*\\[(.+?)\\]\\(([^)]+)\\)[\\s*_]*$")
    ptn_image_link = re.compile("\n?!\\[(.+)\\]\\(([^)]+)\\)\n?")
    ptn_named_link = re.compile("\\[(.+?)\\]\\[(.+?)\\]")
    ptn_link = re.compile("\\[(.+?)\\]\\(([^)]+)\\)")
    ptn_name = re.compile("\\s*\\[(.+?)\\]\\s*:\\s*([^\n]+)")
    ptn_paragraph_break = re.compile("(\\s*\n){2,}")

    result = ''
    is_pre = False  # four-space indent
    is_block = False
    is_last_empty = False
    pending_links = []

    def add_pending_link(m):
        pending_links.append((m.group(2), m.group(1)))

    def add_pending_named_link(m):
        pending_links.append(('[]' + m.group(2), m.group(1)))

    def flush_pending_links(result):
        ptn_dest = re.compile("\n\\s*\\[(.+?)\\]\\s*:\\s*([^\n]+)")
        if len(result) and not result.endswith('\n'):
            result += "\n"
        for url, title in pending_links:
            if url.startswith('[]'):
                # Find the matching named link.
                dest = ptn_dest.search(src)
                if dest:
                    url = dest.group(2)
            result += f"=> {url} {title}\n"
        pending_links.clear()
        return result

    for line in src.split('\n'):
        if not (is_pre or is_block):
            if line == '```':
                is_block = True
                result += "\n```\n"
                continue
            if len(line) == 0:
                is_last_empty = True
                continue
            if is_last_empty:
                result += "\n\n"
            if line[:1] == '#':
                result = flush_pending_links(result)
                line += "\n"
            elif line[:1] == '>':
                result += "\n"
                line += "\n"
            elif line[:1] == '*':
                line += "\n"
            elif len(line) >= 2 and line[:1] in string.digits and \
                 (line[1] == '.' or
                  (line[1] in string.digits and line[2] == '.')):
                result += "\n\n"
                line += "\n"
            elif line[:1] == '|' and line[-1:] == '|':
                line += "\n"
            elif len(result) > 0 and not result[-1] in string.whitespace:
                result += " "
            is_last_empty = False
        elif is_block:
            if not result.endswith('\n'):
                result += '\n'
            if line == '```':
                is_block = False
                result += "```\n"
            else:
                result += line
            continue
        if line[:4] == '    ':
            line = line[4:]
            if not is_pre:
                result += "```\n"
                is_pre = True
        elif is_pre:
            # The space-indentation has ended.
            if not result.endswith('\n'):
                result += '\n'
            result += "```\n"
            if line == "```":
                line = ''
            is_pre = False
        if is_pre:
            result += line + "\n"
        else:
            ln = line
            ln = ptn_name.subn("", ln)[0]
            ln = ptn_standalone_link.subn("\n=> \\2 \\1", ln)[0]
            ln = ptn_image_link.subn("\n=> \\2 \\1\n", ln)[0]
            ln = sub_cb(ptn_named_link, "\\1", ln, add_pending_named_link)
            ln = sub_cb(ptn_link, "\\1", ln, add_pending_link)
            ln = ln.replace("\\_", "_")  # escaped underscores
            result += ln

    result = flush_pending_links(result)
    result = ptn_paragraph_break.subn("\n\n", result)[0]

    return result



# Copyright (c) 2022 Jaakko Keränen <jaakko.keranen@iki.fi>
# License: BSD-2-Clause

"""Static files"""

import fnmatch
import os.path
import string

from gmcapsule import get_mime_type
from pathlib import Path

META = '.meta'
CONFIG = None


def check_meta_rules(path, hostname):
    root = CONFIG.root_dir() / hostname
    dir = Path(path).parent
    while True:
        if not str(dir).startswith(str(CONFIG.root_dir())):
            break
        if (dir / META).exists():
            for rule in open(dir / META, 'rt').readlines():
                rule = rule.strip()
                if len(rule) == 0: continue
                pos = rule.find(':')
                if pos < 0: continue
                rule_path = dir / rule[:pos].strip()
                rule_meta = rule[pos + 1:].strip()
                if fnmatch.fnmatch(root / path, rule_path):
                    if len(rule_meta) >= 4 and rule_meta[2] in string.whitespace:
                        return int(rule_meta[:2]), rule_meta[3:]
                    return 20, rule_meta
        dir = dir.parent

    return 20, get_mime_type(path)


def serve_file(req):
    if req.scheme != 'gemini':
        return 59, "Only Gemini requests allowed"

    cfg = CONFIG
    if req.path == '':
        return 31, '/'

    for seg in req.path.split('/'):
        if seg != '.' and seg != '..' and seg.startswith('.'):
            return 51, "Not found"

    host_root = cfg.root_dir() / req.hostname
    path = os.path.normpath(host_root / req.path[1:])
    if not path.startswith(str(host_root)):
        return 51, "Not found"

    if os.path.isdir(path):
        if not req.path.endswith('/'):
            return 31, req.path + '/'
        path = str(Path(path) / 'index.gmi')

    status, meta = check_meta_rules(path, req.hostname)
    if status and status != 20:
        return status, meta

    if not os.path.exists(path):
        return 51, "Not found"

    # Note: We return a Path object so the sender doesn't have to buffer
    # the entire file in memory first.
    return status, meta, (Path(path) if status == 20 else None)


def init(context):
    cfg = context.config()
    global CONFIG
    CONFIG = cfg
    if 'static' in cfg.ini and 'root' in cfg.section('static'):
        context.print('  Content directory:', cfg.root_dir() / '{hostname}')
        context.add('/*', serve_file)
